#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  Copyright 2020 by leodflag

"""
TimeSeriesAnalysis
=====

module
---------
Chart
    Time series data related drawing function.

Model
    This module contains time series models for forecasting.

ProcessData
    This module contains the access and processing methods of time series data.

"""
